#include <stdint.h>

#define HDRLEN 20
#define START_SEQUENCE 0x12345678

#define ADON_JTAG 0x14
#define ADON_PGMC 0x15
#define ADON_LIOC 0x05

#define IOCTL_READ 0x01
#define IOCTL_WRITE 0x02
#define IOCTL_ACMD 0x03

#define REG_MCSR        0x48            /* Master Control Status */
#define REG_IOC         0x54            /* IOC Data */
#define REG_OMB1        0x60            /* Output Mailbox #1 */
#define REG_OMB2        0x64            /* Output Mailbox #2 */
#define REG_OMB3        0x68            /* Output Mailbox #3 */
#define REG_OMB4        0x6C            /* Output Mailbox #4 */
#define REG_IMB1        0x70            /* Input Mailbox #1 */
#define REG_IMB2        0x74            /* Input Mailbox #2 */
#define REG_IMB3        0x78            /* Input Mailbox #3 */
#define REG_IMB4        0x7C            /* Input Mailbox #4 */

#define MCSR_IMB1F      0x00100000      /* Input Mailbox 1 Full */
#define REG_IOC_DATA    0x00000000
#define REG_IOC_ADDR    0x10000000

struct IceACMD_t {
  uint32_t sync;
  int32_t type;
  int32_t arg1;
  int32_t arg2;
  int32_t arg3;
  int32_t arg4; 
  int32_t arg5; 
}__attribute__((packed));

struct IceJtag_t {
  uint32_t sync;
  uint32_t type;
  uint32_t tms;
  int32_t  tdi;
  uint32_t mask; 
}__attribute__((packed));

#define MAX_LENGTH 64
char g_buffer[MAX_LENGTH];

void setupSerial() {
  SerialUSB.begin(2500000);
  while (!SerialUSB);
}

void processMessage (IceJtag_t* header) {
  boolean rw;
  int i,j,result=0,length;
  int16_t idat;
  IceACMD_t* acmd = (IceACMD_t*)header;

  // handle message types
  if (header->type == ADON_JTAG) {
    switch (header->mask&0xf) {
      case 0xf: length = 32; break;
      case 0x7: length = 24; break;
      case 0x3: length = 16; break;
      case 0x1: length = 8;  break;
      default:  length = 0;
    }
    if (header->mask&0xf00==0) jtag_tms_tdi (header->tms, header->tdi, length);
    else result = jtag_tms_tdi_get (header->tms, header->tdi, length);
  }
  else if (header->type == ADON_PGMC && header->tdi > 0) {
    result = header->tdi; rw = false;
  }
  else if (header->type == ADON_PGMC && header->tdi < 0) {
    result = -header->tdi; rw = true;
  }
  else if (header->type == ADON_LIOC) {
    result = header->tdi; rw = false;
  }
  else if (header->type == IOCTL_WRITE) {
    result = jtag_pci_wr (acmd->arg1,acmd->arg2);
  }
  else if (header->type == IOCTL_READ) {
    result = jtag_pci_rd (acmd->arg1);
  }
  else if (header->type == IOCTL_ACMD) {
    SerialUSB.readBytes((char*)&acmd->arg4,4);    // two more args on ACMDs
    SerialUSB.readBytes((char*)&acmd->arg5,4);    // two more args on ACMDs
    jtag_pci_wr (REG_IMB1,acmd->arg5);
    jtag_pci_wr (REG_OMB4,acmd->arg4);
    jtag_pci_wr (REG_OMB3,acmd->arg3);
    jtag_pci_wr (REG_OMB2,acmd->arg2);
    jtag_pci_wr (REG_OMB1,acmd->arg1);
    for (i=0; i<40; i++) {
      j=jtag_pci_rd(REG_MCSR);
      if ((j&MCSR_IMB1F)!=0) break;
      if (i>10) delay(10);
    }
    if (i>=40) result = 0xFFFF9999;
    else result = jtag_pci_rd(REG_IMB1);
  }
  else { // report error status
    result = header->sync;
  }

  // send message ack
  SerialUSB.write((char*)&result, 4);
  SerialUSB.flush();

  // handle appended bulk programming blocks for JTAG one word at a time
  if (header->type == ADON_PGMC) {
    length = result;
    for (i=0; i<length; i++) {
      if (rw) {
        if (i==0) {
	  // first word of read may have some setup TDI bits
	  SerialUSB.readBytes((char*)&result,4);    
	  result = jtag_tdi_get (result, 32);
	} else {
	  // host wrote just the 1st word - all other TDI=0
	  result = jtag_get (32);
	}
	SerialUSB.write((char*)&result, 4);
      }
      else {
        SerialUSB.readBytes((char*)&result,4);    
	jtag_tdi (result, 32);
      }
    }
    if (rw) SerialUSB.flush();
  }

  // handle appended bulk programming blocks for JVM one word at a time
  if (header->type == ADON_LIOC) {
    length = result*2;
    for (i=0; i<length; i++) {
      SerialUSB.readBytes((char*)&idat,2);    
      jtag_pci_wr (REG_IOC,REG_IOC_ADDR|((idat>>8)&0x3));
      jtag_pci_wr (REG_IOC,REG_IOC_DATA|(idat&0xFF));
    }
  }

}

void processSerial() {
  IceJtag_t* header = (IceJtag_t*) g_buffer;

  // sync to start sequence
  header->sync = 0;
  while (header->sync != START_SEQUENCE) {
    if (!SerialUSB.available()) continue;
    header->sync = header->sync >> 8;
    g_buffer[3] = (char) SerialUSB.read();
  }
  // read rest of message header
  header->type = 0;
  SerialUSB.readBytes(g_buffer+4,HDRLEN-4);
  processMessage(header);
}

