/*****************************************************************************/
/*                                                                           */
/* ICE-SPAC Firmware                                                         */
/*                                                                           */
/* Board: SparkFun SAMD21 Dev Breakout                                       */
/*                                                                           */
/* Revision:                                                                 */
/*   2018 - Baseline                                                         */
/*                                                                           */
/*****************************************************************************/

/*----------------------------- INCLUDES ------------------------------------*/
#include <Arduino.h>
#include <Wire.h>

#define SerialT SerialUSB

#define TCK PORT_PA19
#define TMS PORT_PA18
#define TDO PORT_PA20
#define TDI PORT_PA21

#include "JtagHandlers.h"
#include "SerialHandlers.h"

// Pin Definitions
enum PINS {
  I2C_SDA                = 20,  // PA22
  I2C_SCL                = 21,  // PA23
  USB_N                  = 28,  // PA24
  USB_P                  = 29,  // PA25
  EXTERNAL_BIT0          = 3,   // PA09 INPUT:  FFU
  EXTERNAL_BIT1          = 1,   // PA10 INPUT:  FFU
  EXTERNAL_BIT2          = 4,   // PA08 INPUT:  FFU
  EXTERNAL_BIT3          = 0,   // PA11 INPUT:  FFU
  Q7_SUSPEND_S3_N        = 2,   // PA14 INPUT:  Q7 Suspend S3 Status Pin, actuve low
  Q7_SUSPEND_S5_N        = 5,   // PA15 INPUT:  Q7 Suspend S5 Status Pin, actuve low
  Q7_SUSPEND_STAT_N      = 19,  // PB02 INPUT:  Q7 Suspend Status Pin, actuve low
  Q7_WAKE_N              = 11,  // PA16 INPUT:  Q7 Wake Status Pin, actuve low
  CPU_STDBY_EN           = 17,  // PA04 OUTPUT: 5V Power Rail Enable to 5V Standby on Q7
  MAIN_CURRENT_SENSE     = 18,  // PA05 INPUT:  Tray Current Sense
  ARD_TCK                = 12,  // PA19 OUTPUT: JTAG TCK Outout from Crossbar JTAG Bus
  ARD_TMS                = 10,  // PA18 OUTPUT: JTAG TMS Output from Crossbar JTAG Bus
  ARD_TDO                = 6,   // PA20 INPUT:  JTAG TDO Input from Crossbar JTAG Bus
  ARD_TDI                = 7,   // PA21 cOUTPUT: JTAG TDI OUTPUT from Crossbar JTAG Bus
  ARD_JTAG_EN            = 23,  // PB10 OUTPUT: Arduino Crossbar JTAG Enable, actuve low
  UNDER_OVER_VOLTAGE     = 8,   // PA06 INPUT:  Under / Overvoltage Indicator
  OVER_TEMPERATURE       = 9,   // PA07 INPUT:  Over Voltage Indicator
  FRONT_BUTTON           = 14,  // PA02 INPUT:  Front Switch Button, actuve low
  PM0_QSFP_MODULE_SELECT = 15,  // PB08 OUTPUT: PM0 I2C Select Pin
  PM1_QSFP_MODULE_SELECT = 16,  // PB09 OUTPUT: PM1 I2C Select Pin
  PM2_QSFP_MODULE_SELECT = 13,  // PA17 OUTPUT: PM2 I2C Select Pin
  PM_JTAG_EN             = 38   // PA13 OUTPUT: Header PM JTAG Enable, actuve low
};

// I2C Definititions
enum {
  GPIOID     = 0x20,  // Used for Power Sequencing
  FRONT_TEMP = 0x49,  // Between IOM / K8
  BACK_TEMP  = 0x48   // By Arduino
};

// Class variables
int icnt=0;	// 0.1 sec loop count
int ucnt=0;	// up time in sec
int bcnt=0;	// button press count
int wcnt=0;	// watchdog count
int csense=0;	// current sense
float apower=0;	// average power
float ftemp=0;	// front temperature
float btemp=0;	// back temperature
int qsus=1;	// Q7 suspend state
int pstate=1;	// power state
char pciena='T'; // PCIE enable
char monitor='F'; // monitor mode

unsigned long time,timeup=0;

void getQSus() {
  qsus = 0;
  if (CPAC) return;
  if (digitalRead(Q7_SUSPEND_S3_N)==LOW) qsus |= 0x1;
  if (digitalRead(Q7_SUSPEND_S5_N)==LOW) qsus |= 0x2;
  if (digitalRead(Q7_SUSPEND_STAT_N)==LOW) qsus |= 0x4;
}
void write_gpio_expander(short addr, short val) {
  Wire.beginTransmission(GPIOID);
  Wire.write(addr);
  Wire.write(val);
  Wire.endTransmission();
  return;
}

void setup_gpio_expander() {
  //Sets up Expanders for Output
  Wire.beginTransmission(GPIOID);
  Wire.write(0x01);
  Wire.write(0xFF);//Sets all output states to 0 (off)
  Wire.endTransmission();
  delay(10);
  Wire.beginTransmission(GPIOID);
  Wire.write(0x03);
  Wire.write(0x00);//Sets all pins to outputs
  Wire.endTransmission();
}

/*
  GPIO Enable Bits
  bit[0] =  12V Main Power Enable (Pulled LOW on PCB)
  bit[1] =  5V Q7 Main CPU Power Enable (Pulled LOW on PCB)
  bit[2] =  2.5V / 3.3V / 5.0V Power Enable (Pulled LOW on PCB)
  bit[3] =  12V K8 Power Enable (Pulled LOW on PCB)
  bit[4] =  Q7 Power Power Good Pin (Pulled HIGH on Q7)
  bit[5] =  Q7 Reset Button (Pulled HIGH on Q7)
  bit[6] =  Q7 Power Button (Pulled HIGH on Q7)
  bit[7] =  Q7 Sleep Button (Pulled HIGH on Q7)
*/
void powerOff() {
  SerialT.println("Powering Unit Off");
  write_gpio_expander(0x01, 0xB0);
  digitalWrite(PM2_QSFP_MODULE_SELECT, LOW);
  digitalWrite(PM1_QSFP_MODULE_SELECT, LOW);
  digitalWrite(PM0_QSFP_MODULE_SELECT, LOW);
  pstate=0;
}

void powerOn() {
  SerialT.println(" ");
  digitalWrite(PM2_QSFP_MODULE_SELECT, HIGH);
  digitalWrite(PM1_QSFP_MODULE_SELECT, HIGH);
  digitalWrite(PM0_QSFP_MODULE_SELECT, HIGH);
  delay(1000);
  write_gpio_expander(0x01, 0xBF);	// powerBricks=On powerButton=Off
  delay(1000);
  write_gpio_expander(0x01, 0xFF);	// powerBricks=On powerButton=On
  delay(1000);
  SerialT.println("Powering Unit On...");
  pstate=1;
  ucnt=0;
  write_gpio_expander(0x01, 0xBF);	//turn toggle Power Button Low for 2 seconds
  delay(2000);
  write_gpio_expander(0x01, 0xFF);	//turn on Power Button
  SerialT.println("Power Sequence Finished");
}

void powerLow() {
  SerialT.println("Entering Low Power State");
  delay(1000);
  pstate=2;
  write_gpio_expander(0x01, 0xF0);//turn off all but standby 5V
  digitalWrite(PM2_QSFP_MODULE_SELECT, LOW);
  digitalWrite(PM1_QSFP_MODULE_SELECT, LOW);
  digitalWrite(PM0_QSFP_MODULE_SELECT, LOW);
  SerialT.println("Press Front Button or send WOL packet");
}

void power (char* state) {
  if (strstr(state,"on")!=NULL) powerOn();
  if (strstr(state,"off")!=NULL) powerOff();
}

void resetQ7() {
  write_gpio_expander(0x01, 0xDF);
  delay(100);
  write_gpio_expander(0x01, 0xFF);
}

float cur2pow (int itot, int n) {
  float pavg = 0.0;
  // Compute average of ADC readings
  pavg = (n > 0) ? (n) : (1);
  pavg = ((float)itot) * (1.0 / pavg);
  pavg = (pavg >= 0.0) ? (pavg) : (0.0);
  // Formula for converting current value to power in Watts
  //    3.3 = ADC reference voltage
  // 2860.0 = K (ST Micro chip conversion constant)
  //   12.0 = DC operating voltage of the switch
  // 1000.0 = Sense resistor value (1000 ohms)
  // 1023.0 = ADC normalization (11 bits positive scale)
  pavg = (pavg * 3.3 * 2860.0 * 12.0) / (1000.0 * 1024.0);
  return pavg;
}

float tempSensorRead (int id) {
  int temp=0, status=0;
  Wire.beginTransmission(id);
  Wire.write(0x00);
  status = Wire.endTransmission();
  if (status!=0) return -1;
  status = Wire.requestFrom(id, 2);
  if (status != 2) return -2;
  while (Wire.available()) { 
    status = Wire.read();
    temp = (temp<<8) | (status&0xFF);	// MSB first
  }
  return (float)temp/256;	// 12bit A2D left justified to signed 16b in C
}

void sendStatus() {
  SerialT.print(HW_DEVICE_NAME); 
  SerialT.print(" State="); SerialT.print(pstate);
  SerialT.print(" Q7Sus="); SerialT.print(qsus);
  SerialT.print(" Power="); SerialT.print(apower);
  SerialT.print(" TempF="); SerialT.print(ftemp);
  SerialT.print(" TempB="); SerialT.print(btemp);
  SerialT.print(" UpTime="); SerialT.print(ucnt);
  SerialT.print("\n");
}

void getCommand() {
  int i;
  for (i=0; i<MAX_LENGTH && SerialT.available(); i++) {
    g_buffer[i] = (char)SerialT.read();
    if (i==3 && *(int*)g_buffer==START_SEQUENCE) {
      SerialT.readBytes(g_buffer+4,HDRLEN-4);
      digitalWrite(ARD_JTAG_EN, LOW);
      processMessage((IceJtag_t*)g_buffer);
      digitalWrite(ARD_JTAG_EN, HIGH);
      i-=4; // check for next message immediately
    }
  }
  if (i==0) return;
  SerialT.write("CMD="); SerialT.write(g_buffer,i-1); SerialT.write("\n");
  SerialT.flush();
       if (strncmp(g_buffer,"status",6)==0) sendStatus();
  else if (strncmp(g_buffer,"monitor=",8)==0) monitor=g_buffer[8];
  else if (strncmp(g_buffer,"pciena=",7)==0) pciena=g_buffer[7]; 
  else if (strncmp(g_buffer,"power=",6)==0) power(g_buffer+6); 
  else SerialT.write("Unsupported Command\n");
  SerialT.write("endCMD\n"); 
  SerialT.flush();
}

void setup() {

  // Open serial port, sets data rate to 9600 bps
  Serial.begin(2500000);

  // Open I2C bus
  Wire.begin();

  // Initialize Digital Pin I/O configuration
  pinMode(EXTERNAL_BIT3, INPUT);
  pinMode(EXTERNAL_BIT1, INPUT);
  pinMode(EXTERNAL_BIT0, INPUT);
  pinMode(EXTERNAL_BIT2, INPUT);

  pinMode(Q7_SUSPEND_S3_N, INPUT);
  pinMode(Q7_SUSPEND_S5_N, INPUT);
  pinMode(Q7_WAKE_N, INPUT);
  pinMode(Q7_SUSPEND_STAT_N, INPUT);

  pinMode(ARD_TCK, OUTPUT); digitalWrite(ARD_TCK, LOW);
  pinMode(ARD_TMS, OUTPUT); digitalWrite(ARD_TMS, LOW);
  pinMode(ARD_TDO, INPUT);
  pinMode(ARD_TDI, OUTPUT); digitalWrite(ARD_TDI, LOW);
  pinMode(ARD_JTAG_EN, OUTPUT); digitalWrite(ARD_JTAG_EN, HIGH);

  pinMode(UNDER_OVER_VOLTAGE, INPUT);
  pinMode(OVER_TEMPERATURE, INPUT);

  pinMode(PM0_QSFP_MODULE_SELECT, OUTPUT); digitalWrite(PM0_QSFP_MODULE_SELECT, HIGH);
  pinMode(PM1_QSFP_MODULE_SELECT, OUTPUT); digitalWrite(PM1_QSFP_MODULE_SELECT, HIGH);
  pinMode(PM2_QSFP_MODULE_SELECT, OUTPUT); digitalWrite(PM2_QSFP_MODULE_SELECT, HIGH);

  pinMode(FRONT_BUTTON, INPUT_PULLUP);
  pinMode(MAIN_CURRENT_SENSE, INPUT);
  pinMode(CPU_STDBY_EN, OUTPUT); digitalWrite(CPU_STDBY_EN, HIGH);

  // Setup GPIO Expander
  setup_gpio_expander();

  // Turn on the power to everything
  powerOn();
}

void loop() {

  do {                          // 0.1 second poll on Serial port command buffer
    getCommand();
    time = millis();
  }
  while (time-timeup<100);

  csense += analogRead(MAIN_CURRENT_SENSE);

  int fb = digitalRead(FRONT_BUTTON);
  if (fb==0 && pstate>0 && ++bcnt==30) powerOff();
  if (fb==0 && pstate==0 && ++bcnt==3) powerOn();
//  if (fb==1 && bcnt>2 && bcnt<10 && pstate==1) resetQ7();
  if (fb==1) bcnt=0;

  timeup = time;
  if (++icnt<10) return;	// 1 sec poll

  ucnt++;			// upTime
  apower = cur2pow(csense,10);
  ftemp = tempSensorRead (FRONT_TEMP);
  btemp = tempSensorRead (BACK_TEMP);

  getQSus();

  csense=0;
  icnt=0;

  if (pstate==1 && qsus!=0 && ucnt<20) powerOn();	// redo powerOn
  if (pstate==1 && qsus==7 && ucnt>20) powerOff();	// soft powerOff
  if (pstate==2 && (qsus&0x2)==0) powerOn();		// wake on LAN

  if (monitor!='T') return;     // not producing monitor output

  sendStatus();

}

