/*****************************************************************************/
/*                                                                           */
/* ICE-uPAC Firmware                                                         */
/*                                                                           */
/* Board: Arduinoe-uPAC DevIce                                               */
/*                                                                           */
/* This is C++ code after arduiono preprocess from .ino                      */
/*                                                                           */
/* Revision:                                                                 */
/*   2018 - Baseline                                                         */
/*                                                                           */
/*****************************************************************************/

// Includes
#include <Arduino.h>
#include <Wire.h>	// Wire becomes static class instance
#include <SPI.h>	// SPI becomes static class instance
#include <SD.h>		// SD becomes static class instance

Sd2Card sdc;
SdVolume sdv;
SdFile sdr,sdd,sdf;

#define TCK PORT_PA19
#define TMS PORT_PA18
#define TDI PORT_PA21
#define TDO PORT_PA20

#define SerialT SerialUSB

#include "JtagHandlers.h"
#include "SerialHandlers.h"

// Pin Definitions
enum PINS {
  I2C_SDA                = 20,    // PA22
  I2C_SCL                = 21,    // PA23
  USB_N                  = 28,    // PA24
  USB_P                  = 29,    // PA25
  EXT_JTAG_EN_N          = 0,     // PA11 OUTPUT: Cable JTAG enable,, active low
  ARD_JTAG_EN_N          = 1,     // PA10 OUTPUT: Arduino JTAG enable, active low
  ARD_TCK                = 12,    // PA19 OUTPUT: JTAG TCK Outout from Crossbar JTAG Bus
  ARD_TMS                = 10,    // PA18 OUTPUT: JTAG TMS Output to Crossbar JTAG Bus
  ARD_TDI                = 7,     // PA21 OUTPUT: JTAG TDI OUTPUT to Crossbar JTAG Bus
  ARD_TDO                = 6,     // PA20 INPUT:  JTAG TDO Input from Crossbar JTAG Bus
  MAIN_CURRENT_SENSE     = 14,    // PA02 INPUT:  Tray Current Sense
  ARD_12V_EN             = 15,    // PB08 OUTPUT: 12V Power Enable, active high
  PCIE_RST_N             = 16,    // PB09 OUTPUT: PCIe reset, active low
  USBC1_EN               = 17,    // PA04 OUTPUT: USBC1 switch enable, active high
  USBC2_EN               = 18,    // PA05 OUTPUT: USBC2 switch enable, active high
  GPIO1                  = 30,    // PB22 OUTPUT: GPIO1 to K8M
  GPIO2                  = 31,    // PB23 OUTPUT: GPIO2 to K8M
  QSFP1_MODULE_SELECT    = 3,     // PA09 OUTPUT: QSFP1 I2C Select Pin
  QSFP2_MODULE_SELECT    = 4,     // PA08 OUTPUT: QSFP2 I2C Select Pin
  QSFP1_RESET_N          = 8,     // PA06 OUTPUT: QSFP1 reset, active low
  QSFP2_RESET_N          = 9,     // PA07 OUTPUT: QSFP2 reset, active low
  QSFP1_PRES_N           = 13,    // PA17 INPUT:  QSFP1 present, active low
  QSFP2_PRES_N           = 11,    // PA16 INPUT:  QSFP2 present, active low
  FLASH_MOSI             = 34,    // PA12 OUTPUT: SD Card MOSI
  FLASH_SCK              = 35,    // PA13 OUTPUT: SD Card clock
  FLASH_NSS              = 36,    // PA14 OUTPUT: SD Card chip select, active low
  FLASH_MISO             = 37,    // PA15 INPUT:  SD Card MISO
  FLASH_CD               = 26     // PA27 INPUT:  SD card detect, active low
};

// I2C Address Definititions
enum {
  GPIOID        = 0x20,  // Used for Power Sequencing on SuPAC
  UP_CPS_ID     = 0x30,
  DN_CPS_ID     = 0x32,
  TUSB_ID       = 0x20
};

// static variables
int icnt=0;	// 0.1 sec loop count
int ucnt=0;	// up time in sec
int bcnt=0;	// button press count
int wcnt=0;	// watchdog count
int csense=0;	// current sense
int pstate=1;	// power state
int flash=-2;	// flash detect
float apower=0;	// average power
char pciena='T'; // PCIE enable
char monitor='F'; // monitor mode
int swmode=0;	// switch mode
unsigned long time,timeup=0;

void i2c_write (byte add, byte reg, byte data) {
  Wire.beginTransmission(add);		// 7-bit address, write
  Wire.write(reg);			// register number
  Wire.write(data);			// 8-bit data value
  Wire.endTransmission();
}

byte i2c_read (byte add, byte reg) {
  Wire.beginTransmission(add);		// 7-bit address, write
  Wire.write(reg);			// register number
  Wire.endTransmission();
  Wire.requestFrom(add,1);		// 7-bit address, read 1 byte
  char read_data=reg;
  while (Wire.available()) read_data = Wire.read();	// read data			
  Wire.endTransmission();
  return read_data;
}

void dumpI2Cregs (String text, int addr, int nreg) {
  SerialT.print(text);
  for (byte i=1; i <= nreg; i++) {
    SerialT.print(" 0x");
    char rd = i2c_read (addr, i);
    SerialT.print(rd,HEX);       
  }
  SerialT.print("\n");
}
void getregs() {
  dumpI2Cregs ("UpStream CPS, Reg:",UP_CPS_ID,9);
  dumpI2Cregs ("DnStream CPS, Reg:",DN_CPS_ID,9);
}

/*
    0 : USB 3.1
    1 : USB 3.2
    2 : Display port, lane 0
    3 : Display port, lane 1
    4 : Auxiliary port, USBC serial  
    5 : USB 2.0
    6 : TGbE 
*/

void setupI2Cregs (int mode) {
  SerialT.println("Setup Upstream crosspoint switch registers...");

  // set SWITCH_EN bit
  i2c_write (UP_CPS_ID, 0x01, 0x80);      // Enable switch

 if (mode==0) { // USB 3.1
  i2c_write (UP_CPS_ID, 0x02, 0x01);      // Connect OP1 (RX1) to IP1, 3.1, 3.2, TB
  i2c_write (UP_CPS_ID, 0x04, 0x08);      // Connect OP3 (TX1) to IP4, 3.1, 3.2, TB
 }
 if (mode==1) { // TB, 3.2
   i2c_write (UP_CPS_ID, 0x03, 0x02);   // Connect OP2 (RX2) to IP2
   i2c_write (UP_CPS_ID, 0x05, 0x10);   // Connect OP4 (TX2) to IP5
 }
 if (mode==2 || mode==3) { // DP
   i2c_write (UP_CPS_ID, 0x03, 0x04);   // Connect OP2 (RX2) to IP3
   i2c_write (UP_CPS_ID, 0x05, 0x20);   // Connect OP4 (TX2) to IP6
 }

  // set OP5 CTRL register (Aux)
  i2c_write (UP_CPS_ID, 0x06, 0x40);      // Connect AUX port

  // set SW_CTRL register
  i2c_write (UP_CPS_ID, 0x08, 0x3f);      // enable all bits

  SerialT.println("Setup Downstream crosspoint switch registers...");

  // set SWITCH_EN bit
  i2c_write (DN_CPS_ID, 0x01, 0x80);

 if (mode==0) { // USB 3.1
  i2c_write (DN_CPS_ID, 0x02, 0x01);      // Connect OP1 (RX1) to IP1, 3.1, 3.2, TB
  i2c_write (DN_CPS_ID, 0x04, 0x08);      // Connect OP3 (TX1) to IP4, 3.1, 3.2, TB
 }
 if (mode==1) { // TB, 3.2
   i2c_write (DN_CPS_ID, 0x03, 0x02);   // Connect OP2 (RX2) to IP2
   i2c_write (DN_CPS_ID, 0x05, 0x10);   // Connect OP4 (TX2) to IP5
 }
 if (mode==2 || mode==3) { // DP
   i2c_write (DN_CPS_ID, 0x03, 0x04);   // Connect OP2 (RX2) to IP3
   i2c_write (DN_CPS_ID, 0x05, 0x20);   // Connect OP4 (TX2) to IP6
 }

  // set SOP5 CTRL register (Aux)
  i2c_write (DN_CPS_ID, 0x06, 0x40);      // Connect AUX port

  // set SW_CTRL register
  i2c_write (DN_CPS_ID, 0x08, 0x3f);      // enable all bits
}

float cur2pow (int itot, int n) {
  float pavg = 0.0;
  // Compute average of ADC readings
  pavg = (n > 0) ? (n) : (1);
  pavg = ((float)itot) * (1.0 / pavg);
  pavg = (pavg >= 0.0) ? (pavg) : (0.0);
  // Formula for converting current value to power in Watts
  //    3.3 = ADC reference voltage
  // 2860.0 = K (ST Micro chip conversion constant)
  //   12.0 = DC operating voltage of the switch
  // 1000.0 = Sense resistor value (1000 ohms)
  // 1023.0 = ADC normalization (11 bits positive scale)
  pavg = (pavg * 3.3 * 2860.0 * 12.0) / (1000.0 * 1024.0);
  return pavg;
}

void sendStatus() {
  SerialT.print(HW_DEVICE_NAME); 
  SerialT.print(" State="); SerialT.print(pstate);
  SerialT.print(" Flash="); SerialT.print(flash);
  SerialT.print(" Power="); SerialT.print(apower);
  SerialT.print(" UpTime="); SerialT.print(ucnt);
  SerialT.print("\n");
}

void openSD() {
  SerialT.print("Initializing SD card...");
  if (sdc.init(SPI_HALF_SPEED,SDCARD_SS_PIN) && sdv.init(sdc) && sdr.openRoot(sdv)) {
//  if (sdc.init(SPI_FULL_SPEED,SDCARD_SS_PIN) && sdv.init(sdc) && sdr.openRoot(sdv)) {
    SerialT.println("initialization done.");
    flash=1;
  } else if (flash==-1) {
    SerialT.println("reinitialization ");
    flash=1;
  } else {
    SerialT.println("initialization failed!");
    flash=0;
  }
}
void closeSD() {
  sdr.close();
//  sdv.cacheFlush();
  flash = -1;
}
void checkSD() {
  int cd = digitalRead(FLASH_CD);
  if (cd==0 && flash<0) openSD();
  if (cd==1 && flash>=0) closeSD();
}
void dir (String path) {
  if (flash<=0) { SerialT.print("Flash not present\n"); return; }
  SerialT.print("listDir="); SerialT.print(path); SerialT.print("\n");
  sdr.ls(LS_DATE|LS_SIZE,0);
}
void put (char* path) {
  int blk; char buf[1024];
  sdf.open(&sdr,path,FILE_WRITE);
  sdf.seekSet(0);
  while (true) {
    SerialT.readBytes((char*)(&blk),4);
    if (blk<=0) break;
    blk = SerialT.readBytes(buf,blk);
    if (blk>0) sdf.write(buf,blk);
  }
  sdf.close();
}
void get (char* path) {
  int blk=1024; char buf[1024];
  sdf.open(&sdr,path,FILE_READ);
  sdf.seekSet(0);
  SerialT.readBytes((char*)(&blk),4);	// sync word also max buffer
  while (true) {
    blk = sdf.read(buf,512);
    SerialT.write((char*)(&blk),4);
    if (blk<=0) break;
    blk = SerialT.write(buf,blk);
  }
  sdf.close();
}
void del (char* path) {
  sdf.remove(&sdr,path);
}
void load (char* path) {
  int i,j,blk=512,data; int buf[128];
  i=sdf.open(&sdr,path,FILE_READ);
  SerialT.print("Loading file="); SerialT.print(path); 
  if (i<=0) { SerialT.print(" not found\n"); return; }
  digitalWrite(ARD_JTAG_EN_N, LOW);
  jtag_tms_tdi (JTAG_TMS_CMDR,JTAG_CMD_JSHUTD,32);
  jtag_tms_tdi (JTAG_TMS_STARTUP,0x0,16);
  jtag_tms_tdi (JTAG_TMS_RST,0x0,8);
  jtag_tms_tdi (JTAG_TMS_CMDR,JTAG_CMD_JPROGB,32);
  jtag_tms_tdi (JTAG_TMS_CMD,JTAG_CMD_CFGIN,32);
  udelay(20000);
  SerialT.print(" ");
  for (i=0;;i++) {
    blk = sdf.read((char*)buf,512);
    if (blk<=0) break; blk=blk>>2;
    for (j=0; j<blk; j++) jtag_tdi_bitrev (buf[j],32);
    if (i%100==0) SerialT.print("&\n"); 
  }
  jtag_tms_tdi (JTAG_TMS_SET32X,0x0,32);
  jtag_tms_tdi (JTAG_TMS_SET32,0x0,32);
  jtag_tms_tdi (JTAG_TMS_RST,0x0,8);
  jtag_tms_tdi (JTAG_TMS_CMDR,JTAG_CMD_JSTART,32);
  for (i=0; i<64; i++) jtag_tms_tdi (JTAG_TMS_SET32X,0x0,32);
  jtag_tms_tdi (JTAG_TMS_STARTUP,0x0,16);
  jtag_tms_tdi (JTAG_TMS_RST,0x0,8);
  digitalWrite(ARD_JTAG_EN_N, HIGH);
  sdf.close();
  SerialT.print(" done\n");
}
void setreg (char *path) {
}
void getreg (char *path) {
}
void power (char* state) {
  if (strstr(state,"on")!=NULL) digitalWrite(ARD_12V_EN, HIGH);
  if (strstr(state,"off")!=NULL) digitalWrite(ARD_12V_EN, LOW);
}

void getCommand() {
  int i; char stmp[40];
  for (i=0; i<MAX_LENGTH && SerialT.available(); i++) {
    g_buffer[i] = (char)SerialT.read();
    if (i==3 && *(int*)g_buffer==START_SEQUENCE) {	// binary command from UIC device
      SerialT.readBytes(g_buffer+4,HDRLEN-4);
      digitalWrite(ARD_JTAG_EN_N, LOW);
      processMessage((IceJtag_t*)g_buffer);
      digitalWrite(ARD_JTAG_EN_N, HIGH);
      i-=4; // check for next message immediately
    }
  }
  if (i==0) return;
  SerialT.write("CMD="); SerialT.write(g_buffer,i-1); SerialT.write("\n");	// USB serial command from arduino.c 
  SerialT.flush();
       if (strncmp(g_buffer,"status",6)==0) sendStatus();
  else if (strncmp(g_buffer,"dir",3)==0) dir("/");
  else if (strncmp(g_buffer,"put=",4)==0) put(g_buffer+4);
  else if (strncmp(g_buffer,"get=",4)==0) get(g_buffer+4);
  else if (strncmp(g_buffer,"del=",4)==0) del(g_buffer+4);
  else if (strncmp(g_buffer,"load=",5)==0) load(g_buffer+5);
  else if (strncmp(g_buffer,"power=",6)==0) power(g_buffer+6);
  else if (strncmp(g_buffer,"pciena=",7)==0) { pciena=g_buffer[7]; digitalWrite(PCIE_RST_N, (pciena=='T')?HIGH:LOW); }
  else if (strncmp(g_buffer,"setreg:",7)==0) setreg(g_buffer+7);
  else if (strncmp(g_buffer,"getreg:",7)==0) getreg(g_buffer+7);
  else if (strncmp(g_buffer,"getregs",7)==0) getregs();
  else if (strncmp(g_buffer,"monitor=",8)==0) monitor=g_buffer[8];
  else if (strncmp(g_buffer,"swmode=",5)==0) { swmode=g_buffer[5]-'0'; setupI2Cregs(swmode); }
  else SerialT.write("Unsupported Command\n");
  SerialT.write("endCMD\n");
  SerialT.flush();
}

void setup()
{
  // Open serial port, set data rate
  SerialT.begin(2500000);

  // Open I2C bus
  Wire.begin();

  // Power up K8
  pinMode(ARD_12V_EN, OUTPUT); digitalWrite(ARD_12V_EN, HIGH);
  pinMode(MAIN_CURRENT_SENSE, INPUT);

  // Setup up K8
  pinMode(PCIE_RST_N, OUTPUT); digitalWrite(PCIE_RST_N, (pciena=='T')?HIGH:LOW);
  pinMode(GPIO1, OUTPUT); digitalWrite(GPIO1, LOW);
  pinMode(GPIO2, OUTPUT); digitalWrite(GPIO2, LOW);

  // Setup SSD slot
  pinMode(FLASH_MOSI, OUTPUT); digitalWrite(FLASH_MOSI, LOW);
  pinMode(FLASH_SCK, OUTPUT); digitalWrite(FLASH_SCK, LOW);
  pinMode(FLASH_NSS, OUTPUT); digitalWrite(FLASH_NSS, HIGH);
  pinMode(FLASH_MISO, INPUT);
  pinMode(FLASH_CD, INPUT);
  
  // Setup JTAG bus drivers
  pinMode(EXT_JTAG_EN_N, OUTPUT); digitalWrite(EXT_JTAG_EN_N, HIGH);  // default to cable - tri-state for now
  pinMode(ARD_JTAG_EN_N, OUTPUT); digitalWrite(ARD_JTAG_EN_N, HIGH);  // default to cable - tri-state for now
  pinMode(ARD_TDO, INPUT);
  pinMode(ARD_TDI, OUTPUT); digitalWrite(ARD_TDI, LOW);
  pinMode(ARD_TMS, OUTPUT); digitalWrite(ARD_TMS, LOW);
  pinMode(ARD_TCK, OUTPUT); digitalWrite(ARD_TCK, LOW);

  // Setup QSFP 
  pinMode(QSFP1_MODULE_SELECT, OUTPUT); digitalWrite(QSFP1_MODULE_SELECT, HIGH);
  pinMode(QSFP2_MODULE_SELECT, OUTPUT); digitalWrite(QSFP2_MODULE_SELECT, HIGH);
  pinMode(QSFP1_RESET_N, OUTPUT); digitalWrite(QSFP1_RESET_N, HIGH);
  pinMode(QSFP2_RESET_N, OUTPUT); digitalWrite(QSFP2_RESET_N, HIGH);
  pinMode(QSFP1_PRES_N, INPUT);
  pinMode(QSFP2_PRES_N, INPUT);

  pinMode(USBC1_EN, OUTPUT); digitalWrite(USBC1_EN, HIGH);
  pinMode(USBC2_EN, OUTPUT); digitalWrite(USBC2_EN, HIGH);

  setupJtag();
//  i2c_write (TUSB_ID, 0x1A, 0x0F);      // first I2C access always fails...
//  i2c_write (TUSB_ID, 0x1A, 0x0F);
//  i2c_write (TUSB_ID, 0x1A, 0x0F);      // first I2C access always fails...
  setupI2Cregs(swmode);
}

void loop() {

  do {				// 0.1 second poll on Serial port command buffer
    getCommand();
    time = millis();
  }
  while (time-timeup<100);

  csense += analogRead(MAIN_CURRENT_SENSE);

  timeup = time;
  if (++icnt<10) return;	// 1 sec poll
  icnt=0;

  ucnt++;			// upTime

  apower = cur2pow(csense,10);
  csense = 0;

  checkSD();

  if (monitor!='T') return;	// not producing monitor output

  sendStatus();

  if (ucnt%5==0) {		// every 5 sec
    getregs();
  }

  if (ucnt%10==2) {		// every 10 sec
    if (flash>0) dir("/");
  }

}
