/**********************************************
 ICE Proprietary Software - do NOT disseminate
 **********************************************/
/*
  GrayChip Tuner Controller Elements

  Jeff Schoen
  Innovative Computer Engineering, Inc.
  1/15/2003

  Since HN2TN see TN irdy=H, all data from HN is forced to diff>5 if necessary.
  But R2TN does wait on TN irdy so it will not hit the diff>5 case and always 
  operate on ocycle which is speed.

  The fromram flag selects between these two modes.

  Note: tunercontrol will make sure both PPC and jtag commands multiplex on to the tuner
        control lines but they cannot simultaneously control the same tuners.
        When banked, the front is controlled by JTAG, the back by PPC.

*/
module tunerin (iclk,irst,irdy,iena,isel,idat, oclk,oval,odat, ocycle,ifmt,fromram,ovsr);
  parameter MODE=0;
  input irst, iclk, iena, oclk, ocycle,ifmt,fromram;
  input [7:0] isel;
  input [7:0] ovsr;
  output irdy, oval;
  input [31:0] idat;
  output [15:0] odat;

  reg oena, oval, irdy, ival, irstd, orst, oprime, oprimed;
  reg [8:0] icnt,icnts;
  reg [9:0] ocnt;
  reg icnts1,icnts2,icntsx;

  wire H=1, L=0;

  wire [4:0] diff = icnts[8:4]-ocnt[9:5];

  wire irdy0 = fromram? (diff[4]!=1) : H;
  //wire irdy4 = fromram? (diff[4:3]!=3) : H;
  wire irdy4 = fromram? (diff[4:3]!=3) : (diff[4:2]!=7);
  wire irdy5 = fromram? (diff[4:3]!=3) : (diff[4:2]!=7);

  always @ (posedge iclk) begin
    irstd <= orst;	// make sure icnt resets while orst to prevent spurious oena
    ival <= iena;
    if (irstd) icnt <= 0;
    else if (ival) icnt <= icnt+1;
    if (irst) irdy <= 0; else irdy <= (MODE==5)? irdy5 : (MODE==4)? irdy4 : irdy0;
  end

  always @ (posedge oclk) begin
    icnts1 <= ival; icnts2 <= icnts1;
    icntsx <= (icnts2 && !icnts1) || orst;
    if (icntsx) icnts <= icnt;
  end

  // this circuit bleeds continuous data to the graychips
  wire fstat = (!oprimed)? oprime : (diff[4:3]==3) || (diff[4:3]==2 && !oena) || (diff>1 && ocycle);
  wire oinc;
  always @ (posedge oclk) begin
    orst <= irst;
    if (orst) oprime  <= 0; else oprime <= !oprimed && (oprime || icnts[5]);
    if (orst) oprimed <= 0; else oprimed <= oprimed || ocnt[3] || !fromram;
    if (orst) ocnt    <= 0; else if (oinc) ocnt <= ocnt+1;
  end

  wire [15:0] odata; wire [1:0] osel;

  RAMB16_S18_S36 rama (
    .ADDRA(ocnt[9:0]), .DIA(16'h0000), .DIPA(2'b00), .CLKA(oclk), .ENA(oena),.WEA(L), .SSRA(L), .DOA(odata), .DOPA(osel),
    .ADDRB(icnt[8:0]), .DIB(idat), .DIPB({2{isel[2:1]}}), .CLKB(iclk), .ENB(ival), .WEB(H), .SSRB(L)
  );

generate
if (MODE>=5) begin:m1
  // Input oversampling
  reg fill,filld,ocap,ocyc,opha,ophad;
  reg [7:0] fcnt;
  reg [15:0] odatb;
  wire cycle = (oena||fill);
  wire done = (fcnt<=1);
  wire load = orst || done;
  always @(posedge oclk) begin
    if (orst) oena <= 0; else oena <= fstat && (!cycle || done);
    if (orst) opha <= 0; else opha <= (oena^opha);
    if (orst) ophad<= 0; else ophad<= opha;
    if (orst) ocap <= 0; else ocap <= oena; 
    if (orst) ocyc <= 0; else ocyc <= cycle; 
    if (orst) oval <= 0; else oval <= ocyc; 
    if (orst) fill <= 0; else fill <= cycle && !done;
    if (orst) filld<= 0; else filld <= fill;
    if (load) fcnt <= ovsr; else if (cycle) fcnt <= fcnt-1;
    if (orst) odatb[0] <= 0; else if (ocap) odatb[0] <= osel[0];
    if (filld) odatb[15:1] <= 0; else odatb[15:1] <= !ifmt? odata[15:1] : ophad? {odata[15:8],7'd0} : {odata[7:0],7'd0};
  end
  assign odat = odatb;
  assign oinc = ifmt? (oena&opha) : oena; 
end
if (MODE<5) begin:m0
  always @(posedge oclk) begin
    if (orst) oena <= 0; else oena <= fstat;
    if (orst) oval <= 0; else oval <= oena; 
  end
  assign odat = odata;
  assign oinc = oena;
end
endgenerate

endmodule


module tunercontrol (
  clk,cs,crd,cwr,cpage,caddr,cdata, 
  sclk,srst,scs,saddr,srd,swr,sdata,
  tcclk,tcrd,tcwr,tcpage,tcaddr,tcdata,
  tclk,tpulse1,tpulse2,tpulse3,tpulse4,
  ovsra,ovsrb,ifmta,ifmtb
  );

  input  clk,cs,crd,cwr;
  input  [7:0] cpage,caddr,cdata;
  input  sclk,srst,scs,srd,swr;
  input [31:0] saddr,sdata;

  output  tcclk,tcrd,tcwr;
  output [7:0] tcpage,tcaddr,tcdata;

  input tclk;
  output tpulse1,tpulse2,tpulse3,tpulse4;
  output [7:0] ovsra,ovsrb;
  output ifmta,ifmtb;

  reg sel,tcwr,qrd;
  wire qstat;
  wire [23:0] qout;
  que16xNp #(24) queue (sclk,srst, scs&swr,qrd, {saddr[15:0],sdata[7:0]},qout, qstat);
  
  reg spget;
  reg ifmta,ifmtb;
  reg [2:0] cnt;
  reg [4:0] speed,spacc;
  reg [7:0] ovsra,ovsrb;
  wire rspeed = spget && (saddr[5:4]==0);
  wire rovsra = spget && (saddr[5:4]==1);
  wire rovsrb = spget && (saddr[5:4]==2);
  wire rmodea = spget && (saddr[5:4]==3) && saddr[2];
  wire rmodeb = spget && (saddr[5:4]==3) && saddr[3];
  always @(posedge sclk) begin
    spget <= scs && swr && (saddr[15:8]==0);
    if (rspeed) speed <= sdata[4:0];
    if (rovsra) ovsra <= sdata[7:0];
    if (rovsrb) ovsrb <= sdata[7:0];
    if (rmodea) ifmta <= sdata[0];
    if (rmodeb) ifmtb <= sdata[0];
  end
  always @(posedge sclk) begin
    sel  <= (sel)? (cnt<6) : (qstat && !cs);
    tcwr <= (sel)? (cnt<4) : (cwr && cs);
    if (!sel) cnt <= 0;
    else cnt <= cnt+1;
    qrd <= (cnt==5);
  end

  assign tcclk  = sclk;
  assign tcrd   = (cs && crd);
  assign tcpage = (sel)? qout[23:16] : cpage;
  assign tcaddr = (sel)? qout[15:8] : caddr;
  assign tcdata = (sel)? qout[7:0] : cdata;

  reg [11:0] tcount;
  reg tpulse1,tpulse2,tpulse3,tpulse4,trig2,trig4;
  always @(posedge tclk) begin
    tcount  <= tcount+1;
    tpulse1 <= tcount[3:0]==0;
    trig2   <= tcount[11];
    tpulse2 <= tcount[11] && !trig2;
    tpulse3 <= tcount[2:0]==0;
    spacc   <= spacc + speed;
    trig4   <= spacc[4];
    tpulse4 <= trig4 ^ spacc[4];
  end

endmodule


module tunerbank (
  clk,crd,cwr,caddr,cdata,cdatao, csx,csy, 
  tclk,tpulse1,tpulse2, tena,tdat, rtena,rtdat,
  freqx,fselx,fgntx,fdatx, freqy,fsely,fgnty,fdaty,
  ta,tc,trdx,twrx,trdy,twry,td,tval,tsa,tclkx,tclky,
  tpxv,tpyv,tpxd,tpyd,tpxr,tpyr, test
  );

  parameter ENABLE=1;

  input  clk,crd,cwr;
  input  [7:0] caddr;
  input  [7:0] cdata;
  output [7:0] cdatao;
  input  csx,csy;

  input  tclk,tpulse1,tpulse2,tena,rtena;
  input  [15:0] tdat,rtdat;

  output freqx,freqy;
  output [1:0] fselx,fsely;
  output fgntx,fgnty;
  output [15:0] fdatx,fdaty;

  output [4:0] ta;
  inout  [7:0] tc;
  output trdx, twrx, trdy, twry;
  output [15:0] td;
  output tval, tsa, tclkx, tclky;
  input  tpxv, tpyv;
  input  [3:0] tpxd, tpyd;
  output tpxr,tpyr;

  output [7:0] test;

  wire H=1, L=0;
  wire twr = (!twrx || !twry);

  // control interface
  assign ta = caddr[4:0];
  assign tc = (twr)? cdata: 8'bZ;
  assign cdatao = tc;

  assign trdx = !(crd && !cwr && csx);
  assign trdy = !(crd && !cwr && csy && !csx); // protect bus drivers
  assign twrx = !(cwr && csx && !caddr[7]);
  assign twry = !(cwr && csy && !caddr[7]);

  reg fi32x,fo32x,fi32y,fo32y,tpulse,fbank_,fbank,fdena,fbena,leak;
  reg [3:0] fenax,fenay;
  reg [2:0] tcount;
  reg wrx,wry;

  always @ (posedge clk) begin
    wrx <= (cwr && csx && caddr[7]);
    wry <= (cwr && csy && caddr[7]);
    fenax[0] <= (wrx && cdata[0])? cdata[7] : fenax[0];
    fenax[1] <= (wrx && cdata[1])? cdata[7] : fenax[1];
    fenax[2] <= (wrx && cdata[2])? cdata[7] : fenax[2];
    fenax[3] <= (wrx && cdata[3])? cdata[7] : fenax[3];
    fi32x    <= (wrx && cdata[5])? cdata[7] : fi32x;
    fo32x    <= (wrx && cdata[6])? cdata[7] : fo32x;
    fenay[0] <= (wry && cdata[0])? cdata[7] : fenay[0];
    fenay[1] <= (wry && cdata[1])? cdata[7] : fenay[1];
    fenay[2] <= (wry && cdata[2])? cdata[7] : fenay[2];
    fenay[3] <= (wry && cdata[3])? cdata[7] : fenay[3];
    fi32y    <= (wry && cdata[5])? cdata[7] : fi32y;
    fo32y    <= (wry && cdata[6])? cdata[7] : fo32y;
    fbank_   <= (wrx && cdata[4])? cdata[7] : fbank_;
  end

  reg ftena;
  wire ftval = (fbank)? rtena : tena;
  always @ (posedge tclk) begin
    fbank <= fbank_;
    if (ftval) tcount <= 0;
    else if (tpulse2) tcount <= tcount+1;
    if (ftval) leak <= 0;
    else leak <= (tcount==7) || leak;  // 500uS timer
    tpulse <= (!ftval && leak && tpulse1) || (fbank && !fbena);
    fbena  <= (rtena || fbena) && fbank;
    fdena  <= (fenax!=0 || fenay!=0) && !fbank;
    ftena  <= fdena && (tena || ftena);
  end
  wire fdtena = (fdena && tena) || ftena;
  wire fbtena = (fbank && rtena) || fbena;

  //hmmm what to do about TC SA on 1st clock cycle
  wire testx,testy;

  // present raw data to tuner data input
  oddr btclkx (tclkx,tclk,L,H);  // inverted clock
  oddr btclky (tclky,tclk,L,H);  // inverted clock
  obff btval  (tval, tclk, !(tpulse || (fbank? rtena : tena)));
  obff btsa   (tsa,  tclk, fbtena || (fdtena && tdat[0]) );

  // present C input (usually constant) unless enabled to reduce power
  obffN #(16) btd (td,tclk, (fdena)? {tdat[15:2],L,L} : rtdat);

  // get the tuner output link port data
  tlp tfx (tclk,tpxv,tpxd, fenax,fi32x,fo32x, freqx,fgntx,fselx,fdatx, testx);
  tlp tfy (tclk,tpyv,tpyd, fenay,fi32y,fo32y, freqy,fgnty,fsely,fdaty, testy);
  assign tpxr = !fenax;
  assign tpyr = !fenay;

  wire _tval = tval && !(crd&&cwr);
  wire _tsa = tsa && !(crd&&cwr);

  reg fault;  always @ (posedge tclk) if (!fdtena) fault <= 0; else fault <= fault || tpulse;

//  assign test = {fenax,_tsa,_tval,tena,tdat[12:11],tdat[7:6]};
  assign test = {fault,fenax,fgntx,leak,rtena,fbtena,fbena,fbank};
  
endmodule


// resynchronize tuner oclock&data with internal tclk
module tlport (pclk, pdat, oclk, oval, odat, rst);
  input pclk,rst,oclk;
  input [3:0] pdat;
  output reg oval;
  output [3:0] odat;

  wire iclk;
  IBUF clk (iclk,pclk) /* synthesis syn_noclockbuf=1 */;
  reg [2:0] pcnt,ocnt;

  always @(negedge iclk or posedge rst) begin
    if (rst) pcnt <= 0;
    else pcnt <= pcnt+1;
  end

  //Synthesis attribute ram_style of ram is distributed
  reg [3:0] ram [7:0] /* synthesis syn_ramstyle="select_ram" */;
  always @(negedge iclk) ram[pcnt] <= pdat;
  assign odat = ram[ocnt];

  always @(posedge oclk or posedge rst) begin
    if (rst) ocnt <= 0;
    else if (oval) ocnt <= ocnt+1;
    if (rst) oval <= 0;
    else oval <= pcnt[2] || (ocnt[2:0]!=0 && ocnt[2:0]!=7);
  end

endmodule

// graychip linkport output mode fifo
// this takes exactly 24 slices

module tlp (clk, ival,idat, ena,i32,o32, oval,orst,osel,odat, test);
  input clk, ival, i32, o32;
  input [3:0] idat, ena;
  output oval,orst;
  output [1:0] osel;
  output [15:0] odat;
  output test;

  wire L=0, H=1;

  reg [1:0] tsel;
  reg [2:0] count;
  reg [3:0] itmp, tag, rsel;
  reg primed, irst, iena, gettag;
  wire compress = i32 && !o32;
  wire ready = (count==7);

  // SRL input logic
  always @(posedge clk) begin
    irst <= (ena==0);
    iena <= ival;
    if (irst) count <= 0;
    else if (iena) count <= count+1;
    gettag <= (i32 && count==3) || (!i32 && count[1:0]==1);
    if (gettag) tag <= idat;
    if (ival) itmp <= (gettag) ? {idat[3:2],L,L}: idat;
    if (ready) tsel <= tag[1:0];
    primed <= (ready || primed) && !irst;
    if (!iena) rsel <= rsel+1;
  end
  assign test = (tag==1);

  // SRL word composition and swapping
  reg [15:0] sra;
  always @ (i32 or count) begin
   case ({i32,count})
    4'h0 : sra = 16'h1111;
    4'h1 : sra = 16'h2111;
    4'h2 : sra = 16'h2211;
    4'h3 : sra = 16'h2221;
    4'h4 : sra = 16'h1111;
    4'h5 : sra = 16'h2111;
    4'h6 : sra = 16'h2211;
    4'h7 : sra = 16'h2221;
    4'h8 : sra = 16'h0000;
    4'h9 : sra = 16'h1000;
    4'hA : sra = 16'h1100;
    4'hB : sra = 16'h1110;
    4'hC : sra = 16'h2222;
    4'hD : sra = 16'h3222;
    4'hE : sra = 16'h3322;
    4'hF : sra = 16'h3332;
   endcase
  end

  srl16xN #(4) sr0 (clk, (iena && count[1:0]==3), sra[3:0], itmp, odat[3:0]);
  srl16xN #(4) sr1 (clk, (iena && count[1:0]==2), sra[7:4], itmp, odat[7:4]);
  srl16xN #(4) sr2 (clk, (iena && count[1:0]==1), sra[11:8], itmp, odat[11:8]);
  srl16xN #(4) sr3 (clk, (iena && count[1:0]==0), sra[15:12], itmp, odat[15:12]);

  // SRL output logic
  wire rena; 
  mux4xN rmx (rena, ena[0],ena[1],ena[2],ena[3], rsel[3:2]);

  assign oval = (iena && primed && !(compress && count[2]));
  assign orst = (!iena && !rena);
  assign osel = (iena)? tsel : rsel[3:2];

endmodule

module packetizer (oclk,orst,ordy,oena,osel,odat, iclk, 
  req0,sel0,gnt0,in0, req1,sel1,gnt1,in1, req2,sel2,gnt2,in2, req3,sel3,gnt3,in3, test);

  input oclk,orst,oena,iclk;
  output ordy;
  output [31:0] odat;
  input [15:0] in0,in1,in2,in3;
  input req0,req1,req2,req3;
  input [1:0] sel0,sel1,sel2,sel3;
  input gnt0,gnt1,gnt2,gnt3;
  output [7:0] test;
  output [7:0] osel;

  wire [15:0] in;
  wire ival,irst;
  wire [1:0] chan;
  wire icyc,ocyc;
  reg ordy,odiff,ohold,ostart,ocycle;
  reg [1:0] bank,chip;
  reg [3:0] ochk,ocnt;
  reg [4:0] nobank,obank;

  // cycle through each of 4 chips
  always @ (posedge iclk) begin
    bank <= bank+1;
    chip <= bank;
  end

  // multiplex a possible input word
  mux4xNp #(16) mxin  (in,   in0,in1,in2,in3, bank,iclk);
  mux4xNp #(2) mxch  (chan, sel0,sel1,sel2,sel3, bank,iclk);
  mux4xNp  mxval (ival, req0,req1,req2,req3, bank,iclk);
  mux4xNp  mxrst (irst, gnt0,gnt1,gnt2,gnt3, bank,iclk);
  wire ivr = ival || irst;
  wire [3:0] cca = {chip,chan};

  // access/increment the current channel address
  wire [7:0] addr,addrp1;
  assign addrp1 = (irst)? 0 : addr+1;
  cram16xNs #(8) ra (addr, addrp1, cca, iclk, ivr);
  wire tag = (addr[4:0]==0);
  wire [3:0] tagd = (tag)? cca: in[3:0];

  // prime to get rid of 32x4by transient and cycle in/out
  wire iprime,iprimed,ireset;
  wire iprimew = (addr[7] || iprime) && !irst;
  cram16xNd  rcp (iprime, iprimew, cca, iclk,ivr, ochk,iprimed);
  cram16xNsd rcr (           irst, cca, iclk,ivr, ochk,ireset);
  cram16xNsd rci (      addrp1[5], cca, iclk,ivr, ochk,icyc);

  cram16xNs  rco (ocyc, !ocyc, ochk, oclk, ostart || (ireset && ocyc));

  // decommutator logic
  wire starting = ohold && !ostart && !ocycle;
  always @ (posedge oclk) begin
    odiff <= (ocyc!=icyc) && !odiff && !ohold && iprimed && !ireset;
    ohold <= (odiff || (ohold && !ostart)) && !orst;
    // latch the next output bank address
    if (!(odiff||ohold)) nobank <= {ochk,ocyc};
    // restart search at last channel
    if (starting) ochk <= nobank[4:1];
    else ochk <= ochk+1;
    ostart <= starting;
    ocycle <= (ostart || (ocycle && ocnt<14)) && !orst;
    if (ostart) obank <= nobank;
    if (!oena) ocnt <= 0;
    else ocnt <= ocnt+1; 
    ordy <= ostart || (ocycle && !oena);
  end

  assign osel = {4'h0,obank[4:1]};

  // data ram
  wire H=1, L=0;
  RAMB16_S18_S36 ram (
    .ADDRA({cca,addr[5:0]}), .DIA({in[15:4],tagd}), .DIPA(2'b00), .CLKA(iclk), .ENA(ival), .WEA(H), .SSRA(L), 
    .ADDRB({obank,ocnt}), .DIB(32'h00000000), .DIPB(4'b0000), .CLKB(oclk), .ENB(oena), .WEB(L), .SSRB(L),
    .DOB(odat)
  );

  assign test = {ostart,ocycle,ordy,oena,ocyc,icyc,ival,irst};
//  assign test = 0;
//  assign test = {oena,odat[30:28],ival,in[14:12]};
//  assign test = {ostart,ocycle,ochk[3],ireset,ocyc,icyc,ival,irst};

endmodule
