/**********************************************
 ICE Proprietary Software - do NOT disseminate
 **********************************************/
/*
  MultiChannel Core Skeleton
  Handles individual and banked DMA modes

  Addressing: CCXXaaaa
  CC   = chip select
  XX   = system[7], group[6], channel[5:0]
  aaaa = individual channel

  Jeff Schoen
  Innovative Computer Engineering, Inc.
  11/15/2009

*/

`ifndef MC_ENGINE1
  `define MC_ENGINE1 `MC_ENGINE
`endif
`ifndef MC_ENGINE1S
  `define MC_ENGINE1S `MC_ENGINES
`endif
`ifndef MC_ENGINE2
  `define MC_ENGINE2 `MC_ENGINE
`endif
`ifndef MC_ENGINE2S
  `define MC_ENGINE2S `MC_ENGINES
`endif
`ifndef MC_ENGINE3
  `define MC_ENGINE3 `MC_ENGINE
`endif
`ifndef MC_ENGINE3S
  `define MC_ENGINE3S `MC_ENGINES
`endif
`ifndef MC_ENGINE4
  `define MC_ENGINE4 `MC_ENGINE
`endif
`ifndef MC_ENGINE4S
  `define MC_ENGINE4S `MC_ENGINES
`endif
 
module mcengine (sclks,srst, scs,saddr, swrs,swrbus, srds,srdbus, sack,
	         ioclk, istat,iena,isel,ibus, ostat,oena,osel,obus, test);

  parameter PORT=1;	// engine index
  parameter IBW=64;	// I/O bus width - 32b in processor modules, 64b in main boards

  localparam SBW=32;	// system bus width 
  localparam NC=(PORT==4)?`MC_ENGINE4S : (PORT==3)?`MC_ENGINE3S :  (PORT==2)?`MC_ENGINE2S : `MC_ENGINE1S;	// Number of cores
  localparam IC=1;	// input channels
`ifdef VIRTEX8
  localparam JBW=64;	// internal core I/O bus width 
`else
  localparam JBW=32;	// internal core I/O bus width 
`endif
  localparam JAW=JBW/8;	// internal auxiliary bus width
  localparam IAW=IBW/8;	// I/O auxiliary bus width
  localparam CBW=(NC>16)?5:(NC>8)?4:(NC>4)?3:2;
`ifdef S6M
  localparam FLGO=`PACKET+`OREG;
`else
  localparam FLGO=`PACKET+0;
`endif
  localparam MC=(PORT<=2)?1:0;
  localparam MT=(PORT>=3)?1:0;
`ifdef SYS_CLKF2_MASK
  localparam CLKF2=(`SYS_CLKF2_MASK >> (PORT-1)) & 1;
`else
  localparam CLKF2=0;
`endif


  input [1:0] sclks;	// system clock
  input srst;		// system reset
  input scs;		// system select (qualifies swr and srd for this core)
  input [1:0] swrs;	// register write
  input [1:0] srds;	// register read
  input [31:0] saddr;	// register address
  input [31:0] swrbus;	// register data
  output [31:0] srdbus;	// register data
  output sack;		// acknowledge

  input ioclk;		// I/O data bus clock
  output istat;		// istat - ready to accept 64 by burst
  input iena;		// input enable (one cycle early) (continuous 64 byte bursts)
  input [7:0] isel;	// input channel selection
  input [IBW-1:0] ibus;	// input data bus
  output ostat;		// ostat - ready to provide 64 byte burst
  input oena;		// output enable (one cycle early) (continuous 64 byte bursts)
  output [7:0] osel;	// output channel selection
  output [IBW-1:0] obus;// output data bus

  output [7:0] test;	// 8 test signals for scope probes (test bit=0 of sub-cores must be enable)

  wire L=0, H=1;
  assign sack=H;
  wire sclk = sclks[CLKF2];
  wire swr = swrs[CLKF2];
  wire srd = srds[CLKF2];
  wire ioclk_ = (MC>0)? ioclk : sclk;

  // core variables 
  wire irdy,enad;
  wire [3:0]     sbank = saddr[19:16];		// specific channel bank
  wire [1:0]     gsys  = saddr[23:22];		// system|group addressing
  wire           sys   = saddr[23];		// system addressing
  wire           grp   = saddr[22];		// group addressing
  wire [CBW-1:0] schn  = saddr[CBW-1+16:16];	// specific channel address
  wire [31:0]    saddr_ = {16'b0,saddr[15:0]};
  wire [31:0]    srdbus_[NC-1:0],srdbusg,srdbusc;
  wire [JBW-1:0] jbus,obus_[NC-1:0];
  reg  [JBW-1:0] kbus,ibus_[NC-1:0];
  reg  [3:0]     kchn;
  wire [JAW-1:0] jsel;
  wire [NC-1:0]  istat_,ostat_,swr_,srst_,ienas_;
  reg  [NC-1:0]  ena_;
  reg  [CBW-1:0] wchn;
  reg ival,ena,ienb,ienc,wena,frst,jstat,kstat;

  wire [31:0] tracer;

  // system status ram with parameterized ID 
  swrstatdbg #("MCOR",NC,IC,32'h1CE) ss (sclk,scs,swr&sys, saddr,swrbus,srdbusg, sclk,tracer);

  // global multicore settings
  reg [4:0] mchn;
  reg sbase,echn,split,bulk,ctag,itag,rsas,swrsys,scsc,swrc,swap;
  always @(posedge sclk) begin
    sbase  <= (saddr[10:0]==0);			// the base register of each core
    swrsys <= scs && swr && (gsys>1);
    scsc   <= scs &&        (gsys<2);
    swrc   <= scs && swr && (gsys<2);
    if (swrsys) echn  <= swrbus[1];		// disable startup tags
    if (swrsys) split <= swrbus[2];
    if (swrsys) bulk  <= swrbus[3];		// 512by chunks
    if (swrsys) rsas  <= swrbus[4];
    if (swrsys) swap  <= swrbus[5];
    if (swrsys) ctag  <= swrbus[6];
    if (swrsys) itag  <= swrbus[7];
    if (swrsys) mchn  <= swrbus[20:16];
  end

  // input bus distributor - slow to system clock for easier routing
  always @(posedge ioclk) ival <= iena;
  fifoNxM #(IBW+IAW,JBW+JAW,0,3,`BPAR,0) fi (ioclk,frst, istat,ival,{{4{isel[3:0]}},ibus}, ioclk_,frst, irdy,ienb,{jsel,jbus});
  wire [3:0] jchn = (JAW>4)? {jbus[48],jbus[32],jbus[16],jbus[0]} : {jbus[16],jbus[0]};
  delaypipe #(8) obc (sclk,ena,enad);
  always @(posedge sclk) begin
    ena   <= (ena_!=0);				// any cores enabled
    frst  <= srst || (rsas? ena^enad : ena&!enad);	// reset fifos at restart of channels
  end
  always @(posedge ioclk_) begin
    jstat <= (ena_^istat_)==0;			// cumulative input ready
    ienb  <= irdy && (itag? istat_[jchn] || (jchn>=mchn) : jstat); 			// feed data in 64by bursts
    ienc  <= ienb;				// data available on jbus
    kbus  <= jbus;				// register for broadcast route timing
    kchn  <= jchn;
  end
  wire ichn = jsel[0];
`ifdef ICE_MB
  wire tcen = jsel[1] && !itag;			// use timecode synchronization on bit=0
  wire tsen = jsel[2];				// timecode sync enable
`else
  wire tsen = jsel[1];				// timecode sync enable
  wire tcen = jsel[2] && !itag;			// use timecode synchronization on bit=0
`endif
  wire crdy = (MC>0) || (tcen?tsen:H);

  fanoutp #(NC) foswr (swr_,swr,sclk);
  fanoutp #(NC) fosrst (srst_,srst,sclk);

  // core instantiations 
  genvar i;
  generate
  for (i=0; i<NC; i=i+1) begin:cor
    wire sack_;
    wire odd = (i&1)? H:L;
    wire [7:0] osel_,test_;
    reg  scs_,iena_,oena_,ienb_;
    always @(posedge sclk) begin
      scs_ <= scsc && ( grp? (sbank[0] || sbank[odd+1]) && (!sbank[3] || i<mchn) : (schn==i) ); // channel select by group bits or index
      if (srst) ena_[i] <= 0; else if (swr && scs_ && sbase) ena_[i] <= swrbus[0];	// core enable bit
    end
    always @(posedge ioclk_) begin
      ienb_ <= ena_[i] && (iena_ || ienb_);
      iena_ <= ena_[i] && ienc && (crdy || ienb_) && (!split || !(odd^ichn)) && (!itag || (jchn==i));
      oena_ <= wena && (wchn==i);
      ibus_[i] <= kbus;				// register for broadcast route timing
    end
    assign ienas_[i] = iena_;
    wire [1:0] swrs_ = {2{swr_[i]}};
    if (PORT==4) begin:p4
    `MC_ENGINE4 #(PORT,JBW) engi (sclks,srst_[i], scs_,saddr_, swrs_,swrbus, srds,srdbus_[i], sack_,
	                      ioclk_, istat_[i],iena_,8'h0,ibus_[i], ostat_[i],oena_,osel_,obus_[i], test_);
    end
    if (PORT==3) begin:p3
    `MC_ENGINE3 #(PORT,JBW) engi (sclks,srst_[i], scs_,saddr_, swrs_,swrbus, srds,srdbus_[i], sack_,
	                      ioclk_, istat_[i],iena_,8'h0,ibus_[i], ostat_[i],oena_,osel_,obus_[i], test_);
    end
    if (PORT==2) begin:p2
    `MC_ENGINE2 #(PORT,JBW) engi (sclks,srst_[i], scs_,saddr_, swrs_,swrbus, srds,srdbus_[i], sack_,
	                      ioclk_, istat_[i],iena_,8'h0,ibus_[i], ostat_[i],oena_,osel_,obus_[i], test_);
    end
    if (PORT==1) begin:p1
    `MC_ENGINE1 #(PORT,JBW) engi (sclks,srst_[i], scs_,saddr_, swrs_,swrbus, srds,srdbus_[i], sack_,
	                      ioclk_, istat_[i],iena_,8'h0,ibus_[i], ostat_[i],oena_,osel_,obus_[i], test_);
    end
  end
  endgenerate

  // collector variables
  reg wstr,wenb,wenc,wval,wpha;
  reg [CBW-1:0] wchk, wchnb,wchnc;
  reg [3:0] wchnx; 
  reg [6:0] wcnt; 
  reg [JAW-1:0] wchnv; 
  reg [JBW-1:0] wbus;
  wire [7:0] ochn = wchnc;
  wire ordy = ostat_[wchk];
  wire wrap = (wchk==NC);
  wire wsam = (wchk==wchn) && wena;
  wire wrdy;

  // collector logic
  always @(posedge ioclk_) begin
    if (wrap) wchk <= 0; else if ((!ordy)||wstr||wsam) wchk <= wchk+1;
    if (wstr) wchn <= wchk;
    if (wstr) wpha <= 0; else wpha <= !wpha;
    wstr <= ordy && wrdy && !wstr && !(wena && (bulk? (wcnt<126):(wcnt[3:0]<14)));
    wena <= wstr || (wena && (bulk? (wcnt!=127):(wcnt[3:0]!=15)));
    if (!wena) wcnt <= 0; else wcnt <= wcnt+1;
    wenb  <= wena; wenc <= wenb; wval <= wenc;
    wchnb <= wchn; wchnc <= wchnb; 
    wchnv <= ctag? 0 : (JAW>4)? ochn : wpha? ochn[7:4] : ochn[3:0];
    wchnx <= wchnc;
    wbus  <= obus_[wchnc];
  end
  wire [JBW-1:0] wbusx = (JAW>4 && ctag)? {wbus[63:49],wchnx[3], wbus[47:33],wchnx[2], wbus[31:17],wchnx[1], wbus[15:1],wchnx[0]} 
                       : (JAW<5 && swap)? {wbus[23:16],wbus[31:24],wbus[7:0],wbus[15:8]} : wbus;

//  fifoNxM #(JBW+JAW,IBW+IAW,0,4,`BPAR,FLGO) fo (ioclk_,frst, wrdy,wval,{wchnv,wbusx}, ioclk,frst, ostat,oena,{osel,obus});
  fifoNxM #(JBW+JAW,IBW+IAW,8,4,`BPAR,FLGO) fo (ioclk_,frst, wrdy,wval,{wchnv,wbusx}, ioclk,frst, ostat,oena,{osel,obus});

  reg c0; always @(posedge sclk) c0 <= grp || (schn==0);
`ifdef MC_COMSTAT
  // common status ram for write/readback only status needs (256B*16 = 4kB)
  spram #(4,32,0) rs (sclk,scsc, {grp? 4'd0:schn,saddr[7:2]},swrc, swrbus,srdbusc);
  assign srdbus = sys? srdbusg : c0? srdbus_[0] : srdbusc;
`else
  // mux of individual core status busses
  reg [CBW-1:0] rchn; always @(posedge sclk) if (c0) rchn <= 0; else rchn <= schn;
  assign srdbusc = srdbus_[rchn];
  assign srdbus = sys? srdbusg : srdbusc;
`endif
 
//  assign test = test_[0];
//  assign test = {wchk[2:0],ordy,wstr,wena,oena,iena};
//  assign test = {wval,tsen,tcen,ienb_[0],iena_[0],ienb,iena,srst};
//  assign test = {jstat,istat,ostat,ena,oena,wval,ienb,iena};
//  assign test = {istat,ostat,iena,oena, wchk[1:0],ordy,wval};
  assign test = {istat,ostat,iena,oena, ienas_[1:0],irdy,ienb};
//  assign test = {ena_[1],ena_[0],istat,wval,crdy,tcen,ienb,iena};
//  assign tracer = {istat,ostat,iena,oena,wchn[1:0],wchk[1:0],ostat_[3:0],ordy,wpha,wval,wstr};
//  assign tracer = {istat,ostat,iena,oena,wchn[1:0],wchk[1:0],ostat_[3],ostat_[2],ostat_[1],ostat_[0],ordy,wpha,wval,wstr};
  assign tracer = {istat,ostat,iena,oena,wchn[1:0],wchk[1:0],wval,wrdy, ostat_[1:0],istat_[1:0],ienas_[1:0],jchn[3:0],ienb,irdy};

endmodule

/*
    wstr
    0 wena 0  wchn
    1 wenb 1  wchnb
    0 wenc 2  wchnc
    1 wval 3  wchnv

      wena 14
      wena 15


    ienb
    ienc  jbus
          kbus  iena_
	        ibus_

*/
