/**********************************************
 ICE Proprietary Software - do NOT disseminate
 **********************************************/
/*
  WafeForm Generator Core 

  Jeff Schoen
  Innovative Computer Engineering, Inc.
  1/15/2003

*/
module waveengine (sclk,srst, scs,saddr, swr,swrbus, srd,srdbus, sack,
		   ioclk, istat,iena,isel,ibus, ostat,oena,osel,obus, test);

  parameter PORT=1;	// engine index
  parameter IBW=64;	// I/O bus width - 32b in processor modules, 64b in main boards (PIC5)

  localparam SBW=32;	// System bus width - always 32
  localparam DBW=32;	// Data bus width (internal core use)

  input sclk;		// system clock (typically 100-150MHz)
  input srst;		// system reset
  input scs;		// system select (qualifies swr and srd for this core)
  input swr;		// register write
  input srd;		// register read
  input [31:0] saddr;	// register address
  input [31:0] swrbus;	// register data
  output [31:0] srdbus;	// register data
  output sack;          // acknowledge

  input ioclk;          // I/O data bus clock
  output istat;         // istat - ready to accept 64 by burst
  input iena;           // input enable (one cycle early) (continuous 64 byte burst)
  input [7:0] isel;     // input channel selection
  input [IBW-1:0] ibus; // input data bus
  output ostat;         // ostat - ready to provide 64 byte burst
  input oena;           // output enable (one cycle early) (continuous 64 byte burst)
  output [7:0] osel;    // output channel selection
  output [IBW-1:0] obus;// output data bus

  output [7:0] test;	// 8 test signals for scope probes
  
  wire L=0, H=1;
  assign sack=H;
  assign osel=0;	// single channel default

  // system status ram with parameterized ID 
  swrstat #("WAVE",0) ss (sclk,scs,swr, saddr, swrbus,srdbus);

  // System register enables - with routing pipeline
  wire [7:0] sb; swrbank #(0,8) swb (sclk,scs,swr,saddr,sb);
  wire swrsys  = sb[0]; // system register (required)
  wire swrdec  = sb[1]; // decimation register (optional)
  wire swrgain = sb[2]; // gain register (optional)
  wire swrfreq = sb[3]; // freq register (optional)
  wire swrflag = sb[7]; // flag register (optional)

  // system register - written to by pic_cport() host control software
  reg [31:0] system,dec,gain;	
  always @(posedge sclk) begin
    if (swrsys)  system <= swrbus;
    if (swrdec)  dec    <= swrbus;
    if (swrgain) gain   <= swrbus;
  end
  wire enable     = system[0];		// core enable
  wire dir        = system[1];		// direction ? output:input
  wire reduce     = system[2];		// reduction ? enable:disable
  wire [3:0] ifmt = system[11:8];	// input format  bit[2:0]? 0=16b 1=8b 2=4b 3=1b 4=32b  bit[3]?complex:real
  wire [3:0] ofmt = system[15:12];	// output format bit[2:0]? 0=16b 1=8b 2=4b 3=1b 4=32b  bit[3]?complex:real

  wire reset = srst || !enable;		// core reset (at boot or disabled)
  wire vena,wena,vstat,wstat;
  wire [DBW-1:0] vbus,wbus;

  reg ival; always @(posedge ioclk) ival <= iena; // handle iena early

  // this reformats any input type to output format 4'h8 = complex 16b
  fifoNxMfmt #(IBW,DBW) fi (ioclk,reset, ifmt,istat,ival,ibus, sclk,reset, 4'h8, vstat,vena,vbus);
  assign vena = L;

  // example function - replace with call to user verilog or vhdl here
  wavefunc inst (sclk,reset,reduce, scs,swr,saddr,swrbus, wstat,wena,wbus);

  // this reformats the 4'h8 = complex 16b to any output type 
  fifoNxMfmti #(IBW,DBW) fo (ioclk,reset, ofmt,ostat,oena,obus, sclk,reset, 4'h8, wstat,wena,wbus);

  assign test = {oena,ostat,iena,istat,system[3:0]};	// example test probe

endmodule


// this function should be more generic without ICE dataflow 
module wavefunc (clk,reset,reduce, scs,swr,saddr,swrbus, wstat,wena,wbus);
  input clk,reset,reduce;
  input scs,swr;
  input [31:0] saddr,swrbus;
  input wstat;
  output wena;
  output [31:0] wbus;

  // example register setting - internal
  wire swrdec  = scs && swr && (saddr[5:2]==1);	// decimation register
  wire swrrat  = scs && swr && (saddr[5:2]==3);	// sample rate register
  reg [9:0] dec;
  reg [31:0] rate;
  always @(posedge clk) begin
    if (swrdec) dec  <= swrbus[9:0];
    if (swrrat) rate <= swrbus;
  end

  // example noop with decimation
  reg [9:0] count;
  reg [15:0] ramp;
  reg [31:0] wbus;
  reg vena,wena;
  wire zero = (count==0);
  always @(posedge clk) begin
         if (reset) count <= 0; else if (zero) count <= dec[9:0]; else if (vena) count <= count-1;
         if (reset) ramp <= 0;  else if (vena) ramp <= ramp+1;
    vena <= wstat;
    wena <= vena && (zero || !reduce);
    wbus <= {ramp,ramp};
  end

endmodule
